/*!
 * © 2021 David Vines
 *
 * Creative Commons License
 *
 * This work is licensed under a Creative Commons Attribution-NonCommercial-ShareAlike 4.0 International License.
 */

/* globals Names, ROT, OpeningMove, Continuation */
/* exported Rikishi */
class Rikishi {
	constructor(r) {
		this._name = (r && r._name) || Names.getAName();
		this._rank = (r && r._rank) || "";
		this._int = (r && r._int) || ROT.RNG.getPercentage();
		this._spd = (r && r._spd) || ROT.RNG.getPercentage();
		this._agl = (r && r._agl) || ROT.RNG.getPercentage();
		this._str = (r && r._str) || ROT.RNG.getPercentage();
		this._wgt = (r && r._wgt) || ROT.RNG.getPercentage();
		this._bal = (r && r._bal) || ROT.RNG.getPercentage();
		if (r) {
			this._wins = r._wins;
			this._losses = r._losses;
			this._played = [];
			this._previousBashos = r._previousBashos || [];
		} else {
			this._previousBashos = [];
			this.prepBasho();
		}
		this._cachedSo = undefined;
	}
	getRank() {
		return this._rank;
	}
	setRank(rank) {
		this._rank = rank;
		this._cachedSo = undefined;
	}
	getIntelligence() {
		return this._int;
	}
	getSpeed() {
		return this._spd;
	}
	getAgility() {
		return this._agl;
	}
	getStrength() {
		return this._str;
	}
	getWeight() {
		return this._wgt;
	}
	getBalance() {
		return this._bal;
	}
	hasPlayed(rikishi) {
		return this._played.includes(rikishi);
	}
	played(rikishi) {
		this._played.push(rikishi);
	}
	prepBasho() {
		this._played = [];
		this._wins = 0;
		this._losses = 0;
		this._cachedSo = undefined;
	}
	recordBasho() {
		this._previousBashos.push(this._wins);
	}
	getLastThreeBashoWins() {
		const last3 = this._previousBashos.slice(-3);
		let total = 0;
		for(let wins of last3) {
			total += wins;
		}
		return total;
	}
	getLastTwoBashoWins() {
		const last2 = this._previousBashos.slice(-2);
		let total = 0;
		for(let wins of last2) {
			total += wins;
		}
		return total;
	}
	flattened() {
		let data = {};
		data._name = this._name;
		data._rank = this._rank;
		data._int = this._int;
		data._spd = this._spd;
		data._agl = this._agl;
		data._str = this._str;
		data._wgt = this._wgt;
		data._bal = this._bal;
		data._wins = this._wins;
		data._losses = this._losses;
		data._previousBashos = this._previousBashos;
		data._played = [];
		for(let opp of this._played) {
			data._played.push(opp.toString());
		}
		return data;
	}
	_getMove(them,options) {
		// Build the specific score matrix for this matchup
		let matrix = [];
		for(let i=0; i<options.length; i++) {
			matrix[i] = [];
			for(let j=0; j<options.length; j++) {
				matrix[i][j]  = options[i].getOdds(options[j],this,them)*(103-(them.getBalance())/3);
				matrix[i][j] -= options[j].getOdds(options[i],them,this)*(103-(this.getBalance())/3);
			}
		}
		let oddments = new Array(options.length).fill(0);
		let rowtotal = new Array(options.length).fill(0);
		let coltotal = new Array(options.length).fill(0);
		let ourmove = Math.floor(ROT.RNG.getUniform()*options.length);
		for(let i=0; i<this.getIntelligence()-1; i++) {
			oddments[ourmove] += 1/this.getIntelligence();
			for(let j=0; j<options.length; j++) {
				coltotal[j] += matrix[ourmove][j];
			}
			let lowest = coltotal[0], lowcol = [0];
			for(let j=1; j<coltotal.length; j++) {
				if (coltotal[j] < lowest) {
					lowest = coltotal[j]; lowcol = [j];
				} else if (coltotal[j] === lowest) {
					lowcol.push(j);
				}
			}
			const theirmove = ROT.RNG.getItem(lowcol);

			for(let j=0; j<options.length; j++) {
				rowtotal[j] += matrix[j][theirmove];
			}
			let highest = rowtotal[0], highrow = [0];
			for(let j=1; j<rowtotal.length; j++) {
				if (rowtotal[j] > highest) {
					highest = rowtotal[j]; highrow = [j];
				} else if (rowtotal[j] === highest) {
					highrow.push(j);
				}
			}
			ourmove = ROT.RNG.getItem(highrow);
		}
		oddments[ourmove] += 1/this.getIntelligence();

		// And make the choice
		let roll = ROT.RNG.getUniform(), choice = 0;
		while(roll > oddments[choice]) {
			roll -= oddments[choice++];
		}
		return options[choice];
	}
	getOpeningMove(them) {
		return this._getMove(them,[OpeningMove.Charge,OpeningMove.Attack,OpeningMove.StepAside,OpeningMove.Slap]);
	}
	getContinuation(them) {
		return this._getMove(them,[Continuation.Lever,Continuation.Lift,Continuation.Push,Continuation.Pull,Continuation.Trip]);
	}
	saves() {
		return (ROT.RNG.getPercentage() < (this._bal+3)/3);
	}
	toString() {
		return this._name;
	}
	getScore() {
		return this._wins+"-"+this._losses;
	}
	addWin() {
		this._wins++;
		this._cachedSo = undefined;
	}
	addLoss() {
		this._losses++;
		this._cachedSo = undefined;
	}
	getWins() {
		return this._wins;
	}
	getLosses() {
		return this._losses;
	}
	getSortOrder() {
		if (!this._cachedSo) {
			this._cachedSo = this._wins*1e7 - this._losses*1e5;
			if (this._rank === "Yokozuna") { this._cachedSo += 600; }
			if (this._rank === "Ozeki") { this._cachedSo += 500; }
			if (this._rank === "Ozeki (±)") { this._cachedSo += 400; }
			if (this._rank === "Sekiwake") { this._cachedSo += 300; }
			if (this._rank === "Komusubi") { this._cachedSo += 200; }
			if (this._rank.startsWith("Maegashira ")) { this._cachedSo += (100-Number(this._rank.slice(11))); }
			if (this._rank === "Juryo") {this.cacheSo -= 100; }
		}
		return this._cachedSo;
	}
}